
/**
* @license
* author: Nikolay Ryabov
* flame-chart-js v3.1.1
* Released under the MIT license.
*/

import { jsx } from 'react/jsx-runtime';
import { useRef, useCallback, useEffect } from 'react';
import { FlameChart, FlameChartContainer } from './index.js';
import useResizeObserver from 'use-resize-observer';

const FlameChartComponent = (props) => {
    const boxRef = useRef(null);
    const canvasRef = useRef(null);
    const flameChart = useRef(null);
    useResizeObserver({
        ref: boxRef,
        onResize: ({ width = 0, height = 0 }) => { var _a; return (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.resize(width, height - 3); },
    });
    const initialize = useCallback(() => {
        const { data, marks, waterfall, timeseries, settings, colors, plugins, timeframeTimeseries } = props;
        if (canvasRef.current && boxRef.current) {
            const { width = 0, height = 0 } = boxRef.current.getBoundingClientRect();
            canvasRef.current.width = width;
            canvasRef.current.height = height - 3;
            flameChart.current = new FlameChart({
                canvas: canvasRef.current,
                data,
                marks,
                waterfall,
                timeseries,
                timeframeTimeseries,
                settings,
                colors,
                plugins,
            });
        }
    }, []);
    const setBoxRef = useCallback((ref) => {
        const isNewRef = ref !== boxRef.current;
        boxRef.current = ref;
        if (isNewRef) {
            initialize();
        }
    }, []);
    const setCanvasRef = useCallback((ref) => {
        const isNewRef = ref !== canvasRef.current;
        canvasRef.current = ref;
        if (isNewRef) {
            initialize();
        }
    }, []);
    useEffect(() => {
        var _a;
        if (props.data) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.setNodes(props.data);
        }
    }, [props.data]);
    useEffect(() => {
        var _a;
        if (props.marks) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.setMarks(props.marks);
        }
    }, [props.marks]);
    useEffect(() => {
        var _a;
        if (props.waterfall) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.setWaterfall(props.waterfall);
        }
    }, [props.waterfall]);
    useEffect(() => {
        var _a;
        if (props.timeseries) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.setTimeseries(props.timeseries);
        }
    }, [props.timeseries]);
    useEffect(() => {
        var _a;
        if (props.timeframeTimeseries) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.setTimeframeTimeseries(props.timeframeTimeseries);
        }
    }, [props.timeframeTimeseries]);
    useEffect(() => {
        if (props.settings && flameChart.current) {
            flameChart.current.setSettings(props.settings);
            flameChart.current.renderEngine.recalcChildrenSizes();
            flameChart.current.render();
        }
    }, [props.settings]);
    useEffect(() => {
        var _a;
        if (props.position) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.setFlameChartPosition(props.position);
        }
    }, [props.position]);
    useEffect(() => {
        var _a;
        if (props.zoom) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.setZoom(props.zoom.start, props.zoom.end);
        }
    }, [props.zoom]);
    useEffect(() => {
        var _a;
        if (props.onSelect) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.on('select', props.onSelect);
        }
        return () => {
            var _a;
            if (props.onSelect) {
                (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.removeListener('select', props.onSelect);
            }
        };
    }, [props.onSelect]);
    return (jsx("div", { className: props.className, ref: setBoxRef, children: jsx("canvas", { ref: setCanvasRef }) }));
};

const FlameChartContainerComponent = (props) => {
    const boxRef = useRef(null);
    const canvasRef = useRef(null);
    const flameChart = useRef(null);
    useResizeObserver({
        ref: boxRef,
        onResize: ({ width = 0, height = 0 }) => { var _a; return (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.resize(width, height - 3); },
    });
    const initialize = useCallback(() => {
        const { settings, plugins } = props;
        if (canvasRef.current && boxRef.current) {
            const { width = 0, height = 0 } = boxRef.current.getBoundingClientRect();
            canvasRef.current.width = width;
            canvasRef.current.height = height - 3;
            flameChart.current = new FlameChartContainer({
                canvas: canvasRef.current,
                settings,
                plugins,
            });
        }
    }, []);
    const setBoxRef = useCallback((ref) => {
        const isNewRef = ref !== boxRef.current;
        boxRef.current = ref;
        if (isNewRef) {
            initialize();
        }
    }, []);
    const setCanvasRef = useCallback((ref) => {
        const isNewRef = ref !== canvasRef.current;
        canvasRef.current = ref;
        if (isNewRef) {
            initialize();
        }
    }, []);
    useEffect(() => {
        var _a;
        if (props.settings) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.setSettings(props.settings);
        }
    }, [props.settings]);
    useEffect(() => {
        var _a;
        if (props.zoom) {
            (_a = flameChart.current) === null || _a === void 0 ? void 0 : _a.setZoom(props.zoom.start, props.zoom.end);
        }
    }, [props.zoom]);
    return (jsx("div", { className: props.className, ref: setBoxRef, children: jsx("canvas", { ref: setCanvasRef }) }));
};

export { FlameChartComponent, FlameChartContainerComponent };
